/*
 * form.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifndef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#include "ScrollText.h"

#include "WWWP.h"

typedef struct _input_state
{
  /* generic stuff */
  int id;                  /* input type */
  Widget w;                /* the widget */
  MLTag *p;                /* input tag */
  MLTag *formp;            /* form tag */
  WWWWidget rw;            /* the head honcho */

  /* text, password stuff */
  Widget tw;               /* asciitext widget from scrollingtext widget */

  /* select stuff */
  int option_count;        /* number of options */
  char **options;          /* option text */
  char **values;           /* optional values for options */
  int selected;            /* selected by option tag */
  Widget *sme;             /* SmeBSB widgets for selection, WWWGetMem */
} InputState;

typedef struct _radio_state
{
  char *name;              /* WWWGetMem */
  Widget w;
  struct _radio_state *next;
} RadioState;              /* WWWGetMem */

typedef struct _form_state
{
  MLTag *p;                /* form tag info */
  MLTag *sp;               /* select tag */
  MLTag *tap;              /* textarea tag */
  int option_count;        /* number of options */
  char **options;          /* option text */
  char **values;           /* optional values for options */
  int selected;            /* selected option */
  RadioState *rslist;      /* radio widgets, allocated by WWWGetMem */
  char *tatext;            /* textarea default text */
  char *otext;             /* option text */
} FormState;               /* allocated by WWWGetMem */

#define INPUT_TEXT 0
#define INPUT_PASSWORD 1
#define INPUT_CHECKBOX 2
#define INPUT_RADIO 3
#define INPUT_SUBMIT 4
#define INPUT_RESET 5
#define INPUT_RANGE 6
#define INPUT_AUDIO 7
#define INPUT_FILE 8
#define INPUT_SCRIBBLE 9
#define INPUT_HIDDEN 10
#define INPUT_IMAGE 11
#define INPUT_TEXTAREA 12
#define INPUT_SELECT 13

/*
 *
 * Private functions
 *
 */

static RenderPart *CreateInputState _ArgProto((WWWWidget, MLTag *,
					       int, int, int));
static void DestroyInputState _ArgProto((InputState *));
static RenderPart *CreateHidden _ArgProto((WWWWidget, MLTag *));
static RenderPart *CreateText _ArgProto((WWWWidget, MLTag *, int));
static RenderPart *CreateCheckbox _ArgProto((WWWWidget, MLTag *));
static RenderPart *CreateRadio _ArgProto((WWWWidget, MLTag *));
static RenderPart *CreateCommand _ArgProto((WWWWidget, MLTag *, int));
static char **AddEntry _ArgProto((char **, int, char *));
static char *NameValueToURLEncoded _ArgProto((char **, char **, int));

/*
 * NameValueToURLEncoded
 */
static char *
NameValueToURLEncoded(names, values, count)
char **names;
char **values;
int count;
{
  int i;
  char *finfo = NULL;
  char *sep = "";
  char *n, *v;
  int flen, alen;
  char *format = "%s%s=%s";

  finfo = NULL;
  flen = 0;
  for (i = 0; i < count; i++)
  {
    if (names[i] == NULL) continue;
     
    n = EscapeURL(names[i], 0);
     
    if (values[i] != NULL) v = EscapeURL(values[i], 1);
    else v = alloc_string("");
      
    alen = flen + strlen(n) + strlen(v) + strlen(sep) + strlen(format) + 1;

    if (finfo != NULL) finfo = realloc_mem(finfo, alen);
    else finfo = alloc_mem(alen);
    sprintf (finfo + flen, format, sep, n, v); 

    sep = "&";
    flen = strlen(finfo);

    free_mem(n);
    free_mem(v);
  }

  return(finfo);
}

/*
 * SubmitCallback
 */
static void
SubmitCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  RenderPart *rp = (RenderPart *)cldata;
  InputState *ci = (InputState *)GetRenderState(rp);
  InputState *xi;
  int i, j;
  char **names;
  char **values;
  char *n, *v;
  Boolean checked;
  char *url;
  URLParts *up;
  char *method;
  char *enctype;
  int count;
  URLParts *bup, *tup;

  if ((url = MLFindAttribute(ci->formp, "action")) == NULL) return;
  if ((up = ParseURL(url)) == NULL) return;
  method = MLFindAttribute(ci->formp, "method");
  enctype = MLFindAttribute(ci->formp, "enctype");

  /*
   * This sucks.
   */
  for (count = 0, rp = ci->rw->www.rphead; rp != NULL; rp = rp->mnext)
  {
    if (rp->p != NULL &&
        (rp->p->id == TAG_INPUT || rp->p->id == TAG_TEXTAREA ||
         rp->p->id == TAG_SELECT))
    {
      xi = (InputState *)GetRenderState(rp);
      if (xi->formp == ci->formp) count++;
    }
  }

  if (count == 0)
  {
    fprintf (stderr, "SubmitCallback error.\n");
    return;
  }

  names = (char **)alloc_mem(sizeof(char **) * count);
  values = (char **)alloc_mem(sizeof(char **) * count);

  for (i = 0, rp = ci->rw->www.rphead; rp != NULL; rp = rp->mnext)
  {
    if (rp->p != NULL &&
	(rp->p->id == TAG_INPUT || rp->p->id == TAG_TEXTAREA ||
	 rp->p->id == TAG_SELECT))
    {
      xi = (InputState *)GetRenderState(rp);
      if (xi->formp == ci->formp)
      {
	names[i] = NULL;
	values[i] = NULL;
	
	if ((n = MLFindAttribute(rp->p, "name")) != NULL)
	{
	  if (xi->id == INPUT_TEXT || xi->id == INPUT_PASSWORD)
	  {
	    names[i] = alloc_string(n);
	    XtVaGetValues(xi->tw, XtNstring, &v, NULL);
	    if (v != NULL) values[i] = alloc_string(v);
	    else values[i] = NULL;
	    i++;
	  }
	  else if (xi->id == INPUT_TEXTAREA)
	  {
	    names[i] = alloc_string(n);
	    XtVaGetValues(xi->w, XtNstring, &v, NULL);
	    if (v != NULL) values[i] = alloc_string(v);
	    else values[i] = NULL;
	    i++;
	  }
	  else if (xi->id == INPUT_CHECKBOX || xi->id == INPUT_RADIO)
	  {
	    XtVaGetValues(xi->w, XtNstate, &checked, NULL); 
	    if (checked)
            {
	      names[i] = alloc_string(n);
	      v = MLFindAttribute(rp->p, "value");
	      if (v != NULL) values[i] = alloc_string(v);
	      else values[i] = NULL;
	      i++;
	    }
	  }
	  else if (xi->id == INPUT_HIDDEN)
	  {
	    v = MLFindAttribute(rp->p, "value");
	    if (v != NULL) values[i] = alloc_string(v);
	    else values[i] = NULL;
	    names[i] = alloc_string(n);
	    i++;
	  }
	  else if (xi->id == INPUT_SELECT)
	  {
	    if (xi->values != NULL && xi->values[xi->selected] != NULL)
	    {
	      values[i] = alloc_string(xi->values[xi->selected]);
	    }
	    else values[i] = alloc_string(xi->options[xi->selected]);
	    names[i] = alloc_string(n);
	    i++;
	  }
	  else if (xi == ci) /* named submit...freak out images probably */
	  {
	    v = MLFindAttribute(rp->p, "value");
	    if (v != NULL) values[i] = alloc_string(v);
	    else values[i] = NULL;
	    names[i] = alloc_string(n);
	    i++;
	  }
	}
      }
    }
  }

  if ((bup = GetBaseURLParts(ci->rw)) != NULL)
  {
    tup = MakeURLParts(up, bup);
    DestroyURLParts(up);
    up = tup;
  }

  up->method = method != NULL ? alloc_string(method):NULL;

  if (enctype == NULL) enctype = "application/x-www-form-urlencoded";
  up->input_data_type = alloc_string(enctype);

  if (i > 0)
  {
    /* Well, only one type used so we might as well go for it. */
    up->input_data = NameValueToURLEncoded(names, values, i);
    up->input_data_len = strlen(up->input_data);
  
    for (j = 0; j < i; j++)
    {
      if (names[j] != NULL) free_mem(names[j]);
      if (values[j] != NULL) free_mem(values[j]);
    }
  }
  else
  {
    up->input_data = alloc_string("");
    up->input_data_len = 0;
  }

  free_mem(names);
  free_mem(values);

  NewDocument(ci->rw, up);
    
  DestroyURLParts(up);

  return;
}

/*
 * ResetCallback
 *
 * Do something about this later.
 */
static void
ResetCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  return;
}

/*
 * SmeCallback
 */
static void
SmeCallback(w, cldata, cbdata)
Widget w;
XtPointer cldata, cbdata;
{
  RenderPart *rp = (RenderPart *)cldata;
  InputState *ci = (InputState *)GetRenderState(rp);
  int i;
  String str;

  for (i = 0; i < ci->option_count; i++)
  {
    if (ci->sme[i] == w)
    {
      ci->selected = i;
      XtVaGetValues(w, XtNlabel, &str, NULL);
      XtVaSetValues(ci->w, XtNlabel, str, NULL);
      return;
    }
  }

  return;
}

/*
 * SetupInput
 */
static void
SetupInput(rw, rp, r, line_y, line_height)
WWWWidget rw;
RenderPart *rp;
Region r;
int line_y;
int line_height;
{
  InputState *ci = (InputState *)GetRenderState(rp);

  if (ci->w == NULL) return;

  XtConfigureWidget(ci->w, (Position)rp->x, (Position)line_y,
		    (Dimension)(rp->width - 10),
		    (Dimension)(rp->height - 5),
		    (Dimension)1);

  XtManageChild(ci->w);

  return;
}

/*
 * CreateInputState
 */
static RenderPart *
CreateInputState(rw, p, width, height, type)
WWWWidget rw;
MLTag *p;
int width, height;
int type;
{
  RenderPart *rp;
  InputState *ci;
  
  rp = CreateRenderPart(rw, p, SetupInput, NULL, DestroyInputState, NULL,
			sizeof(InputState));
  ci = (InputState *)GetRenderState(rp);
  ci->p = p;
  ci->id = type;
  ci->rw = rw;
  ci->w = NULL;
  ci->tw = NULL;
  ci->sme = NULL;
  ci->formp = NULL;
  ci->selected = 0;
  ci->options = NULL;
  ci->values = NULL;
  ci->option_count = 0;

  AddRenderPart(rw, rp, width + 10, height + 5);

  return(rp);
}

/*
 * DestroyInputState
 */
static void
DestroyInputState(ci)
InputState *ci;
{
  int i;

  if (ci->w != NULL) XtDestroyWidget(ci->w);
  if (ci->values != NULL)
  {
    for (i = 0; i < ci->option_count; i++)
    {
      if (ci->values[i] != NULL) free_mem(ci->values[i]); 
    }
    free_mem(ci->values);
  }
  if (ci->options != NULL)
  {
    for (i = 0; i < ci->option_count; i++)
    {
      if (ci->options[i] != NULL) free_mem(ci->options[i]);
    }
    free_mem(ci->options);
  }

  return;
}

/*
 * CreateHidden
 */
static RenderPart *
CreateHidden(rw, p)
WWWWidget rw;
MLTag *p;
{
  RenderPart *rp;

  rp = CreateInputState(rw, p, 0, 0, INPUT_HIDDEN);

  return(rp);
}

/*
 * CreateText
 */
static RenderPart *
CreateText(rw, p, type)
WWWWidget rw;
MLTag *p;
int type;
{
  Widget w, tw;
  int width, height;
  XFontStruct *font;
  char *value;
  RenderPart *rp;
  InputState *ci;
  Boolean echo;
  char *name;

  if (type == INPUT_PASSWORD)
  {
    name = "password";
    echo = False;
  }
  else
  {
    name = "text";
    echo = True;
  }

  w = XtCreateWidget(name,
		     scrollingTextWidgetClass, rw->www.child,
		     NULL, 0);

  XtVaGetValues(w, XtNtextWidget, &tw, NULL);

  if ((value = MLFindAttribute(p, "value")) != NULL)
  {
    value = alloc_string(value);
    WWWSpacify(value, strlen(value));
    XtVaSetValues(tw, XtNeditType, XtEtextEdit, XtNstring, value,
		  XtNecho, echo, NULL);
    free_mem(value);
  }
  else
  {
    XtVaSetValues(tw, XtNeditType, XtEtextEdit, XtNstring, "",
		  XtNecho, echo, NULL);
  }
  
  XtVaGetValues(tw, XtNfont, &font, NULL);
  if (font == NULL) height = 20;
  else height = font->ascent + font->descent + 5;
  width = 100;

  rp = CreateInputState(rw, p, width, height, type);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;
  ci->tw = tw;

  return(rp);
}

/*
 * CreateImage
 */
static RenderPart *
CreateImage(rw, p)
WWWWidget rw;
MLTag *p;
{
  return(NULL);
}

/*
 * CreateCheckbox
 */
static RenderPart *
CreateCheckbox(rw, p)
WWWWidget rw;
MLTag *p;
{
  Widget w;
  Boolean state;
  RenderPart *rp;
  InputState *ci;

  if (MLFindAttribute(p, "checked") != NULL) state = True;
  else state = False;

  w = XtVaCreateWidget("checkbox",
		       toggleWidgetClass, rw->www.child,
		       XtNstate, state,
		       XtNlabel, " ",
		       NULL);
  rp = CreateInputState(rw, p, 15, 15, INPUT_CHECKBOX);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;

  return(rp);
}

/*
 * CreateRadio
 */
static RenderPart *
CreateRadio(rw, p)
WWWWidget rw;
MLTag *p;
{
  Widget w;
  Boolean state;
  FormState *fs = rw->www.formstate;
  RadioState *rs;
  char *name;
  RenderPart *rp;
  InputState *ci;

  if (MLFindAttribute(p, "checked") != NULL) state = True;
  else state = False;

  if ((name = MLFindAttribute(p, "name")) == NULL) name = "";
  for (rs = fs->rslist; rs != NULL; rs = rs->next)
  {
    if (strcasecmp(name, rs->name) == 0) break;
  }

  w = XtVaCreateWidget("radio",
		       toggleWidgetClass, rw->www.child,
		       XtNstate, state,
		       XtNlabel, " ",
		       NULL);
  if (rs == NULL)
  {
    rs = (RadioState *)WWWGetMem(rw, sizeof(RadioState));
    rs->name = (char *)WWWGetMem(rw, strlen(name) + 1);
    strcpy(rs->name, name);
    rs->w = w;
    rs->next = fs->rslist;
    fs->rslist = rs;

    XtVaSetValues(w, XtNstate, True, NULL);
  }
  else XtVaSetValues(w, XtNradioGroup, rs->w, NULL);

  rp = CreateInputState(rw, p, 15, 15, INPUT_RADIO);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;

  return(rp);
}

/*
 * CreateCommand
 */
static RenderPart *
CreateCommand(rw, p, type)
WWWWidget rw;
MLTag *p;
int type;
{
  Arg args[10];
  int argcnt;
  int width, height;
  char *value;
  XFontStruct *font;
  Widget w;
  InputState *ci;
  RenderPart *rp;
  char *name = type == INPUT_SUBMIT ? "submit":"reset";

  argcnt = 0;
  if ((value = MLFindAttribute(p, "value")) != NULL)
  {
    XtSetArg(args[argcnt], XtNlabel, value); argcnt++;
  }
  w = XtCreateWidget(name,
		     commandWidgetClass, rw->www.child,
		     args, argcnt);

  XtVaGetValues(w, XtNlabel, &value, NULL);

  XtVaGetValues(w, XtNfont, &font, NULL);
  if (font == NULL)
  {
    width = 50;
    height = 20;
  }
  else
  {
    width = XTextWidth(font, value, strlen(value)) + 10;
    height = font->ascent + font->descent + 3;
  }
  rp = CreateInputState(rw, p, width, height, type);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;
  
  if (type == INPUT_SUBMIT)
  {
    XtAddCallback(w, XtNcallback, SubmitCallback, (XtPointer)rp);
  }
  else if (type == INPUT_RESET)
  {
    XtAddCallback(w, XtNcallback, ResetCallback, (XtPointer)rp);
  }

  return(rp);
}

Boolean
InOptionEnv(rw)
WWWWidget rw;
{
  RenderEnv *re;

  for (re =  rw->www.restack; re != NULL; re = re->next)
  {
    if (re->tagid == TAG_SELECT) return(False);
    else if (re->tagid == TAG_OPTION) return(True);
  }

  return(False);
}

/*
 * AddEntry
 */
static char **
AddEntry(list, length, entry)
char **list;
int length;
char *entry; /* must be malloc'd */
{
  if (length == 0) list = (char **)alloc_mem(sizeof(char *));
  else
  {
    list = (char **)realloc_mem((char *)list, sizeof(char *) * (length + 1));
  }

  list[length] = entry;

  return(list);
}

/*
 *
 * Public functions
 *
 */

/*
 * HandleForm
 */
void
HandleForm(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  FormState *fs;

  if (end != 0)
  {
    rw->www.formstate = NULL;
    DestroyRenderEnv(rw, TAG_FORM);
  }
  else
  {

    if (rw->www.formstate != NULL) return;

    fs = (FormState *)WWWGetMem(rw, sizeof(FormState));
    fs->p = p;
    fs->sp = NULL;
    fs->tap = NULL;
    fs->option_count = 0;
    fs->options = NULL;
    fs->values = NULL;
    fs->selected = 0;
    fs->rslist = NULL;
    fs->tatext = NULL;
    fs->otext = NULL;
    rw->www.formstate = fs;

    CreateRenderEnv(rw, TAG_FORM);
  }

  return;
}

/*
 * HandleInput
 */
void
HandleInput(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  InputState *ci;
  char *type;
  FormState *fs;
  RenderPart *rp;

  if (end != 0) return;

  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  type = MLFindAttribute(p, "type");
  if (type == NULL || strcasecmp(type, "text") == 0)
      rp = CreateText(rw, p, INPUT_TEXT);
  else if (strcasecmp(type, "checkbox") == 0) rp = CreateCheckbox(rw, p);
  else if (strcasecmp(type, "hidden") == 0) rp = CreateHidden(rw, p);
  else if (strcasecmp(type, "image") == 0) rp = CreateImage(rw, p);
  else if (strcasecmp(type, "password") == 0)
      rp = CreateText(rw, p, INPUT_PASSWORD);
  else if (strcasecmp(type, "radio") == 0) rp = CreateRadio(rw, p);
  else if (strcasecmp(type, "reset") == 0)
      rp = CreateCommand(rw, p, INPUT_RESET);
  else if (strcasecmp(type, "submit") == 0)
      rp = CreateCommand(rw, p, INPUT_SUBMIT);
  else rp = NULL;

  if (rp != NULL)
  {
    ci = (InputState *)GetRenderState(rp);
    ci->formp = fs->p;
  }

  return;
}

void
HandleTextarea(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  FormState *fs;
  InputState *ci;
  int width, height;
  XFontStruct *font;
  char *rstr, *cstr;
  int rows, cols;
  RenderEnv *re;
  char *value;
  Widget w;
  RenderPart *rp;

  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  if (end == 0)
  {
    re = CreateRenderEnv(rw, TAG_TEXTAREA);
    re->text_mode = TAG_TEXTAREA;
    fs->tap = p;
    fs->tatext = NULL;
    return;
  }

  if (fs->tatext == NULL) value = "";
  else value = fs->tatext;

  w = XtVaCreateWidget("textarea",
		       asciiTextWidgetClass, rw->www.child,
		       XtNeditType, XtEtextEdit,
		       XtNstring, value, NULL);

  XtVaGetValues(w, XtNfont, &font, NULL);
  if (font == NULL) height = 20;
  else height = font->ascent + font->descent + 2;
  width = XTextWidth(font, "X", 1);

  if ((rstr = MLFindAttribute(p, "rows")) != NULL)
  {
    rows = atoi(rstr);
    if (rows < 1) rows = 5;
  }
  else rows = 5;
  height *= rows;

  if ((cstr = MLFindAttribute(p, "cols")) != NULL)
  {
    cols = atoi(cstr);
    if (cols < 1) cols = 20;
  }
  else cols = 20;
  width *= cols;

  rp = CreateInputState(rw, fs->tap, width, height, INPUT_TEXTAREA);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;
  ci->formp = fs->p;

  DestroyRenderEnv(rw, TAG_TEXTAREA);

  if (fs->tatext != NULL) free_mem(fs->tatext);

  return;
}

/*
 * HandleSelect
 */
void
HandleSelect(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  FormState *fs;
  InputState *ci;
  int i;
  int width, twidth;
  XFontStruct *font;
  Widget w, smw, *sme;
  RenderPart *rp;

  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  if (end == 0)
  {
    fs->option_count = 0;
    fs->sp = p;

    CreateRenderEnv(rw, TAG_SELECT);
    return;
  }

  if (InOptionEnv(rw)) DestroyRenderEnv2(rw, TAG_OPTION, True);

  w = XtVaCreateWidget("menubutton",
		       menuButtonWidgetClass, rw->www.child,
		       XtNmenuName, "simplemenu",
		       XtNlabel, fs->options[fs->selected],
		       NULL);

  smw = XtVaCreatePopupShell("simplemenu",
			     simpleMenuWidgetClass, w,
			     XtNwidth, 0,
			     XtNheight, 0,
			     NULL);

  sme = (Widget *)WWWGetMem(rw, sizeof(Widget) * fs->option_count);
  XtVaGetValues(w, XtNfont, &font, NULL);
  for (width = 0, i = 0; i < fs->option_count; i++)
  {
    sme[i] = XtVaCreateManagedWidget(fs->options[i],
				     smeBSBObjectClass, smw,
				     NULL);
    
    twidth = XTextWidth(font, fs->options[i], strlen(fs->options[i]));
    if (width < twidth) width = twidth;
  }
  
  DestroyRenderEnv(rw, TAG_SELECT);

  rp = CreateInputState(rw, fs->sp,
			width + 10, font->ascent + font->descent + 4,
			INPUT_SELECT);
  ci = (InputState *)GetRenderState(rp);
  ci->w = w;
  ci->formp = fs->p;
  ci->options = fs->options;
  ci->option_count = fs->option_count;
  ci->values = fs->values;
  ci->selected = fs->selected;
  ci->sme = sme;
  
  for (i = 0; i < fs->option_count; i++)
  {
    XtAddCallback(sme[i], XtNcallback, SmeCallback, (XtPointer)rp);
  }

  return;
}

/*
 * HandleOption
 */
void
HandleOption(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  FormState *fs;
  RenderEnv *re;
  char *value;

  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  if (end == 0)
  {
    if (InOptionEnv(rw)) DestroyRenderEnv2(rw, TAG_OPTION, True);

    if ((value = MLFindAttribute(p, "value")) != NULL)
    {
      value = alloc_string(value);
    }
    fs->values = AddEntry(fs->values, fs->option_count, value);
    value = NULL;

    if (MLFindAttribute(p, "selected") != NULL)
    {
      fs->selected = fs->option_count;
    }

    fs->otext = NULL;

    re = CreateRenderEnv(rw, TAG_OPTION);
    re->text_mode = TAG_OPTION;
    return;
  }

  fs->options = AddEntry(fs->options, fs->option_count, fs->otext);
  fs->otext = NULL;

  WWWSpacify(fs->options[fs->option_count],
	     strlen(fs->options[fs->option_count]));
  fs->option_count++;

  DestroyRenderEnv(rw, TAG_OPTION);

  return;
}

/*
 * TextareaText
 */
void
TextareaText(rw, p)
WWWWidget rw;
MLTag *p;
{
  int len;
  FormState *fs;

  if (p->len == 0) return;
  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  if (fs->tatext == NULL)
  {
    fs->tatext = (char *)alloc_mem(p->len + 1);
    len = 0;
  }
  else
  {
    len = strlen(fs->tatext);
    fs->tatext = (char *)realloc_mem(fs->tatext, len + p->len + 1);
  }

  strncpy(fs->tatext + len, p->text, p->len);
  fs->tatext[len + p->len] = '\0';

  return;
}

/*
 * OptionText
 */
void
OptionText(rw, p)
WWWWidget rw;
MLTag *p;
{
  int len;
  FormState *fs;

  if (p->len == 0) return;
  if ((fs = (FormState *)rw->www.formstate) == NULL) return;

  if (fs->otext == NULL)
  {
    fs->otext = (char *)alloc_mem(p->len + 1);
    len = 0;
  }
  else
  {
    len = strlen(fs->otext);
    fs->otext = (char *)realloc_mem(fs->otext, len + p->len + 1);
  }

  strncpy(fs->otext + len, p->text, p->len);
  fs->otext[len + p->len] = '\0';

  return;
}

