/*
 * layout.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@isri.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Scrollbar.h>

#include "WWWP.h"

static RenderLine *CreateRenderLine _ArgProto((WWWWidget));
static void ConfigureWidgets _ArgProto((WWWWidget));
static void CallTagHandler _ArgProto((WWWWidget, int, MLTag *));

/*
 * WWWSpacify
 *
 * Turns all whitespace to spaces
 */
void
WWWSpacify(s, len)
char *s;
int len;
{
  char *cp, *lastcp = s + len;

  for (cp = s; cp < lastcp; cp++)
  {
    if (isspace(*cp)) *cp = ' ';
  }

  return;
}

/*
 * GetBaseURLParts
 */
URLParts *
GetBaseURLParts(rw)
WWWWidget rw;
{
  if (rw->www.bup != NULL) return(rw->www.bup);
  if (rw->www.dilist == NULL) return(NULL);
  return(rw->www.dilist->up);
}

/*
 * RemainingLineWidth
 */
unsigned int
RemainingLineWidth(rw)
WWWWidget rw;
{
  RenderLine *rl;

  rl = rw->www.rltail;

  return(rl->maxwidth - rl->width);
}

/*
 * GetWidthHeight
 */
int
GetWidthHeight(p, width, height)
MLTag *p;
unsigned int *width, *height;
{
  char *wstr, *hstr;

  wstr = MLFindAttribute(p, "width");
  hstr = MLFindAttribute(p, "height");
  if (wstr != NULL && hstr != NULL)
  {
    *width = (unsigned int)atoi(wstr);
    *height = (unsigned int)atoi(hstr);
    return(0);
  }

  return(-1);
}

void
AddLineBreak(rw)
WWWWidget rw;
{
  CreateRenderLine(rw);

  return;
}

void
AddBlankLine(rw)
WWWWidget rw;
{
  AddLineBreak(rw);
  AddLineBreak(rw);

  return;
}

/*
 * InEnv
 */
bool
InEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  RenderEnv *re;

  for (re = rw->www.restack; re != NULL; re = re->next)
  {
    if (re->tagid == tagid) return(True);
  }

  return(False);
}

/*
 * CreateRenderEnv
 */
RenderEnv *
CreateRenderEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  RenderEnv *re;

  re = (RenderEnv *)WWWGetMem(rw, sizeof(RenderEnv));
  re->tagid = tagid;

  if (rw->www.restack != NULL)
  {
    RenderEnv *ce = rw->www.restack;

    re->anchorPart = ce->anchorPart;
    re->namePart = ce->namePart;
    re->font = ce->font;
    re->align = ce->align;
    re->text_mode = ce->text_mode;
    re->x = ce->x;
    re->width = ce->width;
    re->wrap = ce->wrap;
  }
  else
  {
    re->anchorPart = NULL;
    re->namePart = NULL;
    re->font = rw->www.normalFont;
    re->align = ALIGN_LEFT;
    re->text_mode = TAG_NOTHING;
    re->x = rw->www.margin;
    re->width = rw->www.width;
    re->wrap = True;
  }

  re->next = rw->www.restack;
  rw->www.restack = re;

  return(re);
}

/*
 * DestroyRenderEnv2
 */
void
DestroyRenderEnv2(rw, tagid, call_self)
WWWWidget rw;
int tagid;
bool call_self;
{
  RenderEnv *fe, *te;

  if (rw->www.destroy_locked) return;
  rw->www.destroy_locked = True;

  for (fe = rw->www.restack; fe != NULL; fe = fe->next)
  {
    if (tagid == fe->tagid) break;
  }

  if (fe != NULL)
  {
    for (; rw->www.restack != NULL && rw->www.restack != fe; )
    {
      te = rw->www.restack;
      CallTagHandler(rw, te->tagid, NULL);
      rw->www.restack = rw->www.restack->next;
    }
    if (call_self) CallTagHandler(rw, fe->tagid, NULL);
    rw->www.restack = fe->next;
  }

  rw->www.destroy_locked = False;

  return;
}

/*
 * DestroyRenderEnv
 */
void
DestroyRenderEnv(rw, tagid)
WWWWidget rw;
int tagid;
{
  DestroyRenderEnv2(rw, tagid, False);
  return;
}

/*
 * DestroyRenderPart
 */
void
DestroyRenderPart(rp)
RenderPart *rp;
{
  if (rp->destroy != NULL) (*(rp->destroy))(rp->state);
  return;
}

/*
 * CreateRenderPart
 */
RenderPart *
CreateRenderPart(rw, p, setup, render, destroy, select, state_size)
WWWWidget rw;
MLTag *p;
void (*setup)();
void (*render)();
void (*destroy)();
void (*select)();
int state_size;
{
  RenderPart *rp;

  rp = (RenderPart *)WWWGetMem(rw, sizeof(RenderPart));
  rp->p = p;
  rp->anchorPart = rw->www.restack->anchorPart;
  rp->namePart = rw->www.restack->namePart;

  if (rp->anchorPart != NULL)
  {
    if (rw->www.aphead == NULL) rw->www.aphead = rp;
    else rw->www.aptail->anext = rp;
    rw->www.aptail = rp;
  }

  rp->setup = setup;
  rp->render = render;
  rp->destroy = destroy;
  rp->select = select;
  if (state_size > 0) rp->state = (RenderPart *)WWWGetMem(rw, state_size);
  else rp->state = NULL;

  rp->x = 0;
  rp->y = 0;
  rp->width = 0;
  rp->height = 0;
  rp->breakable = True;
  rp->space = False;
  rp->renderable = False;
  rp->lnext = NULL;
  rp->mnext = NULL;
  rp->anext = NULL;

  /*
   * Add to the master list.
   */
  if (rw->www.rphead == NULL) rw->www.rphead = rp;
  else rw->www.rptail->mnext = rp;
  rw->www.rptail = rp;

  return(rp);
}

/*
 * ComputeLine
 */
static void
ComputeLine(rl)
RenderLine *rl;
{
  RenderPart *rp;

  rl->width = 0;
  for (rp = rl->rphead; rp != NULL; rp = rp->lnext)
  {
    rp->x = rl->x + rl->width;
    rl->width += rp->width;
    if (rp->height > rl->maxy) rl->maxy = rp->height;
  }

  return;
}

/*
 * AddRenderPart
 */
int
AddRenderPart(rw, rp, width, height)
WWWWidget rw;
RenderPart *rp;
unsigned int width, height;
{
  RenderLine *rl = rw->www.rltail;

  rp->width = width;
  rp->height = height;

  if (rl->rphead != NULL &&
      rl->width + rp->width >= rl->maxwidth &&
      rw->www.restack->wrap)
  {
    RenderPart *xp, *fxp;
    RenderLine *orl;

    for (xp = rl->rptail; xp != NULL && !IsBreakable(xp); xp = xp->lprev)
	;

    if (xp != NULL)
    {
      orl = rl;
      rl = CreateRenderLine(rw);

      if (xp->lnext != NULL)
      {
	fxp = xp->lnext;
	xp->lnext = NULL;
	
	ComputeLine(orl);
	
	for (; IsSpace(fxp); fxp = fxp->lnext)
	    ;
	
	fxp->lprev = NULL;
	rl->rphead = fxp;
	
	for (; fxp != NULL; fxp = fxp->lnext)
	{
	  rl->rptail = fxp;
	}
	
	ComputeLine(rl);
      }
    }
  }
  if (rl->rphead == NULL && IsSpace(rp)) return(0);

  rp->x = rl->x + rl->width;
  rl->width += rp->width;
  if (rp->height > rl->maxy) rl->maxy = rp->height;

  if (rl->rptail != NULL) rl->rptail->lnext = rp;
  else rl->rphead = rp;
  rp->lprev = rl->rptail;
  rp->lnext = NULL;
  rl->rptail = rp;

  if (rw->www.delay)
  {
    rw->www.delay = False;
    WWWLayout(rw, NULL);
  }

  return(0);
}

/*
 * CreateRenderLine
 */
static RenderLine *
CreateRenderLine(rw)
WWWWidget rw;
{
  RenderLine *rl;
  RenderPart *rp;
  int adjustx;
  Region r;
  Position x, y;
  Dimension width, height;
  XRectangle rec;

  rl = (RenderLine *)WWWGetMem(rw, sizeof(RenderLine));
  rl->rphead = NULL;
  rl->rptail = NULL;
  rl->height = 0;
  rl->width = 0;
  rl->next = NULL;
  rl->x = rw->www.restack->x;
  rl->maxwidth = rw->www.restack->width;
  rl->miny = 0;
  rl->maxy = 0;
  rl->align = rw->www.restack->align;

  if (rw->www.rltail != NULL)
  {
    RenderLine *ll = rw->www.rltail;

    ll->next = rl;
    if (ll->width > rw->www.width)
    {
      rw->www.width = ll->width;
      adjustx = 0;
    }

    if (ll->align == ALIGN_CENTER) adjustx = (ll->maxwidth - ll->width) / 2;
    else if (ll->align == ALIGN_RIGHT) adjustx = ll->maxwidth - ll->width;
    else adjustx = 0;

    ll->height = ll->maxy - ll->miny;
    if (ll->miny < 0) ll->y -= ll->miny;

    rl->y = rw->www.lineSpace + ll->y + ll->height;
    rw->www.height = ll->y + ll->height;

    /*
     * Search for the region that is on the screen
     */
    r = XCreateRegion();
    XtVaGetValues(rw->www.child, XtNx, &x, XtNy, &y, NULL);
    XtVaGetValues(rw->www.clip, XtNwidth, &width, XtNheight, &height, NULL);
    rec.x = (short)x;
    rec.y = (short)y;
    rec.width = (unsigned short)width;
    rec.height = (unsigned short)height;
    XUnionRectWithRegion(&rec, r, r);

    for (rp = ll->rphead; rp != NULL; rp = rp->lnext)
    {
      rp->x += adjustx;
      if (rp->setup != NULL) (*(rp->setup))(rw, rp, r, ll->y, ll->height);
    }

    XDestroyRegion(r);
  }
  else
  {
    rw->www.rlhead = rl;
    rl->y = rw->www.margin;
  }

  rw->www.rltail = rl;

  ConfigureWidgets(rw);

  return(rl);
}

/*
 * ConfigureWidgets
 *
 * One day the second argument needs to be removed because there will
 * be multiple layouts with all sorts of weird regions that will have to
 * calculated to determine the size of the document.
 */
static void
ConfigureWidgets(rw)
WWWWidget rw;
{
  Dimension child_width, child_height;
  Dimension clip_width, clip_height;
  Dimension vx, hy;
  float shown;

  if (!XtIsRealized((Widget)rw)) return;

  /*
   * Big fun.
   *
   * This is definitely broken code.
   */
  if (rw->www.width > 0x7ffd) child_width = 0x7ffd;
  else child_width = (Dimension)rw->www.width + rw->www.margin * 2;

  if (rw->www.height > 0x7ffd) child_height = 0x7ffd;
  else child_height = (Dimension)rw->www.height + rw->www.margin * 2;

  clip_width = rw->core.width - rw->www.vert_bar_width;
  clip_height = rw->core.height;

  if (child_width > clip_width)
  {
    hy = rw->www.horiz_bar_height;
    clip_height -= hy;
  }
  else
  {
    hy = 0;
    child_width = clip_width;
  }

  if (child_height > clip_height) vx = rw->www.vert_bar_width;
  else
  {
    vx = 0;
    child_height = clip_height;
  }

  if (child_height > clip_height)
  {
    XtConfigureWidget(rw->www.vert_bar,
		      0, hy,
		      rw->www.vert_bar_width, rw->core.height - hy, 0);
    shown = (float)clip_height / (float)child_height;
    XawScrollbarSetThumb(rw->www.vert_bar, (float)-1.0, shown);
    if (!XtIsManaged(rw->www.vert_bar)) XtManageChild(rw->www.vert_bar);
  }
  else XtUnmanageChild(rw->www.vert_bar);

  if (child_width > clip_width)
  {
    XtConfigureWidget(rw->www.horiz_bar,
		      vx, 0,
		      rw->core.width - vx, rw->www.horiz_bar_height, 0);
    shown = (float)clip_width / (float)child_width;
    XawScrollbarSetThumb(rw->www.horiz_bar, (float)-1.0, shown);
    if (!XtIsManaged(rw->www.horiz_bar)) XtManageChild(rw->www.horiz_bar);
  }
  else XtUnmanageChild(rw->www.horiz_bar);

  XtConfigureWidget(rw->www.clip,
		    rw->www.vert_bar_width, hy,
		    clip_width, clip_height, 0);

  XtResizeWidget(rw->www.child, child_width, child_height, 0);

  return;
}

/*
 * CallTagHandler
 */
static void
CallTagHandler(rw, tag, p)
WWWWidget rw;
int tag;
MLTag *p;
{
  int end;

  if (p == NULL) end = 1;
  else end = p->end;

  if (tag == TAG_EOF) HandleEOF(rw);
  else if (tag == TAG_H1) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H2) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H3) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H4) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H5) HandleHx(rw, tag, p, end);
  else if (tag == TAG_H6) HandleHx(rw, tag, p, end);
  else if (tag == TAG_IMG) HandleImg(rw, p, end);
  else if (tag == TAG_TITLE) HandleTitle(rw, end);
  else if (tag == TAG_HR) HandleHR(rw, p, end);
  else if (tag == TAG_P) HandleP(rw, p, end);
  else if (tag == TAG_UL) HandleList(rw, tag, p, end);
  else if (tag == TAG_OL) HandleList(rw, tag, p, end);
  else if (tag == TAG_MENU) HandleList(rw, tag, p, end);
  else if (tag == TAG_DIR) HandleList(rw, tag, p, end);
  else if (tag == TAG_DL) HandleList(rw, tag, p, end);
  else if (tag == TAG_LI) HandleLI(rw, p, end);
  else if (tag == TAG_DT) HandleDT(rw, p, end);
  else if (tag == TAG_DD) HandleDD(rw, p, end);
  else if (tag == TAG_OL) HandleOL(rw, p, end);
  else if (tag == TAG_BR) HandleBR(rw, p, end);
  else if (tag == TAG_ADDRESS) HandleAddress(rw, p, end);
  else if (tag == TAG_A) HandleA(rw, p, end);
  else if (tag == TAG_PRE) HandlePRE(rw, p, end);
  else if (tag == TAG_LISTING) HandleListing(rw, p, end);
  else if (tag == TAG_PLAINTEXT) HandlePlaintext(rw, p, end);
  else if (tag == TAG_B) HandleFontChange(rw, end, tag);
  else if (tag == TAG_EM) HandleFontChange(rw, end, tag);
  else if (tag == TAG_STRONG) HandleFontChange(rw, end, tag);
  else if (tag == TAG_TT) HandleFontChange(rw, end, tag);
  else if (tag == TAG_CITE) HandleFontChange(rw, end, tag);
  else if (tag == TAG_I) HandleFontChange(rw, end, tag);
  else if (tag == TAG_KBD) HandleFontChange(rw, end, tag);
  else if (tag == TAG_FORM) HandleForm(rw, p, end);
  else if (tag == TAG_INPUT) HandleInput(rw, p, end);
  else if (tag == TAG_TEXTAREA) HandleTextarea(rw, p, end);
  else if (tag == TAG_BASE) HandleBase(rw, p, end);
  else if (tag == TAG_HEAD) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_HTML) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_BODY) HandleMisc(rw, tag, p, end);
  else if (tag == TAG_BLOCKQUOTE) HandleBlockquote(rw, p, end);
  else if (tag == TAG_OPTION) HandleOption(rw, p, end);
  else if (tag == TAG_SELECT) HandleSelect(rw, p, end);

  return;
}

/*
 * WWWLayout
 */
void
WWWLayout(rw, p)
WWWWidget rw;
MLTag *p;
{
  if (p != NULL)
  {
    if (rw->www.pqhead == NULL) rw->www.pqhead = p;
    else rw->www.pqtail->qnext = p;
    rw->www.pqtail = p;
  }

  for (; !rw->www.delay && (p = rw->www.pqhead) != NULL; )
  {
    rw->www.pqhead = rw->www.pqhead->qnext;

    if (rw->www.restack->tagid == TAG_PLAINTEXT)
    {
      PlainText(rw, p, True);
    }
    else if (p->id == TAG_TEXT)
    {
      switch (rw->www.restack->text_mode)
      {
	case TAG_TEXTAREA:
          TextareaText(rw, p);
	  break;
	case TAG_OPTION:
          OptionText(rw, p);
	  break;
	case TAG_AU:
	  AuthorText(rw, p);
	  break;
	case TAG_TITLE:
          TitleText(rw, p);
	  break;
        case TAG_PRE:
	case TAG_LISTING:
	case TAG_CODE:
	  PlainText(rw, p, False);
	  break;
        default:
	  FillText(rw, p);
      }
    }
    else CallTagHandler(rw, p->id, p);
  }

  return;
}

/*
 * WWWEndLayout
 */
void
WWWEndLayout(rw)
WWWWidget rw;
{
  DestroyRenderEnv(rw, TAG_START);

  if (rw->www.rphead != NULL)
  {
    RenderPart *t, *c;
    
    for (c = rw->www.rphead; c != NULL; )
    {
      t = c;
      c = c->mnext;
      DestroyRenderPart(t);
    }
  }

  if (rw->www.title != NULL) free_mem(rw->www.title);
  if (rw->www.author != NULL) free_mem(rw->www.author);
  if (rw->www.bup != NULL) DestroyURLParts(rw->www.bup);
  if (rw->www.hs != NULL) MLDestroy(rw->www.hs);
  if (rw->www.memstate != NULL) WWWFreeMem(rw);

  WWWInitLayoutVars(rw);

  return;
}

/*
 * InitLayoutVars
 */
void
WWWInitLayoutVars(rw)
WWWWidget rw;
{
  rw->www.aphead = NULL;
  rw->www.aptail = NULL;
  rw->www.rphead = NULL;
  rw->www.rptail = NULL;
  rw->www.rlhead = NULL;
  rw->www.rltail = NULL;
  rw->www.restack = NULL;
  rw->www.pqhead = NULL;
  rw->www.pqtail = NULL;
  rw->www.delay = False;
  rw->www.bup = NULL;
  rw->www.title = NULL;
  rw->www.author = NULL;
  rw->www.formstate = NULL;
  rw->www.memstate = NULL;
  rw->www.hs = NULL;
  rw->www.destroy_locked = False;

  return;
}

/*
 * WWWStartLayout
 */
void
WWWStartLayout(rw, up, raw)
WWWWidget rw;
URLParts *up;
bool raw;
{
  RenderEnv *re;
  Dimension clip_width, clip_height;

  WWWInitLayoutVars(rw);

  rw->www.hs = MLInit(rw, WWWLayout);

  /*
   * Set the initial dimensions of the clip widget (they don't change)
   * and the initial size of the document.
   */
  clip_width = rw->core.width - rw->www.vert_bar_width;
  clip_height = rw->core.height;

  rw->www.width = (unsigned int)(clip_width - rw->www.margin * 2);
  rw->www.height = 0;

  if (XtIsRealized(rw->www.child))
  {
    XClearWindow(XtDisplay(rw->www.child), XtWindow(rw->www.child));

    XtConfigureWidget(rw->www.clip,
		      rw->www.vert_bar_width, 0,
		      clip_width, clip_height, 0);

    XtConfigureWidget(rw->www.child,
		      0, 0,
		      clip_width, clip_height, 0);

    ConfigureWidgets(rw);
  }

  /*
   * Setup environment
   */
  re = CreateRenderEnv(rw, TAG_START);

  /*
   * Fake a plaintext tag if raw text.
   */
  if (raw) HandlePlaintext(rw, NULL, 0);

  /*
   * Intialize line queue
   */
  CreateRenderLine(rw);

  return;
}

