/*
 * text.c
 *
 * Copyright (c) 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for more information.
 *
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include "WWWP.h"

typedef struct _text_state
{
  char *s;
  int len;
  XFontStruct *font;
} TextState;

/*
 *
 * Private functions
 *
 */
static void RenderText _ArgProto((WWWWidget, RenderPart *, XEvent *));
static RenderPart *CreateTextState _ArgProto((WWWWidget,
					      MLTag *p,
					      char *, int,
					      XFontStruct *));
static void SetupText _ArgProto((WWWWidget, RenderPart *, Region, int, int));

/*
 * SetupText
 */
static void
SetupText(rw, rp, r, line_y, line_height)
WWWWidget rw;
RenderPart *rp;
Region r;
int line_y;
int line_height;
{
  rp->y = line_y + line_height - rp->height;
  rp->renderable = True;
  if (XRectInRegion(r, rp->x, rp->y, rp->width, rp->height))
  {
    RenderText(rw, rp, NULL);
  }

  return;
}

/*
 * RenderText
 */
static void
RenderText(rw, rp, xe)
WWWWidget rw;
RenderPart *rp;
XEvent *xe;
{
  Display *dpy = XtDisplay(rw->www.child);
  Window win = XtWindow(rw->www.child);
  TextState *ts = (TextState *)GetRenderState(rp);

  XSetFont(dpy, rw->www.gc, ts->font->fid);

  if (rp->anchorPart != NULL)
  {
    XSetForeground(dpy, rw->www.gc, rw->www.anchorColor);
    XDrawString(dpy, win, rw->www.gc, rp->x, rp->y + ts->font->ascent,
		ts->s, ts->len);
    XDrawLine(dpy, win, rw->www.gc,
	      rp->x, rp->y + rp->height,
	      rp->x + rp->width, rp->y + rp->height);
    XSetForeground(dpy, rw->www.gc, rw->www.foreground);
  }
  else
  {
    XDrawString(dpy, win, rw->www.gc, rp->x, rp->y + ts->font->ascent,
		ts->s, ts->len);
  }

  return;
}

/*
 * CreateTextState
 */
static RenderPart *
CreateTextState(rw, p, s, slen, font)
WWWWidget rw;
MLTag *p;
char *s;
int slen;
XFontStruct *font;
{
  RenderPart *rp;
  TextState *ts;

  rp = CreateRenderPart(rw, p, SetupText,
			RenderText, NULL, NULL, sizeof(TextState));
  ts = (TextState *)GetRenderState(rp);
  ts->s = s;
  ts->len = slen;
  ts->font = font;

  if (s == NULL)
  {
    s = " ";
    slen = 1;
    rp->space = True;
  }
  else rp->breakable = False;

  AddRenderPart(rw, rp,
		XTextWidth(font, s, slen), font->ascent + font->descent);

  return(rp);
}

/*
 *
 * Public functions
 *
 */

/*
 * TitleText
 */
void
TitleText(rw, p)
WWWWidget rw;
MLTag *p;
{
  int len;

  if (p->len == 0) return;
  if (rw->www.title == NULL)
  {
    rw->www.title = (char *)alloc_mem(p->len + 1);
    len = 0;
  }
  else
  {
    rw->www.title = (char *)realloc_mem(rw->www.title,
					strlen(rw->www.title) + p->len + 1);
    len = strlen(rw->www.title);
  }

  strncpy(rw->www.title + len, p->text, p->len);
  rw->www.title[len + p->len] = '\0';

  WWWSpacify(rw->www.title, len + p->len);

  return;
}

/*
 * AuthorText
 */
void
AuthorText(rw, p)
WWWWidget rw;
MLTag *p;
{
  int len;

  if (p->len == 0) return;
  if (rw->www.author == NULL)
  {
    rw->www.author = (char *)alloc_mem(p->len + 1);
    len = 0;
  }
  else
  {
    rw->www.author = (char *)realloc_mem(rw->www.author,
					 strlen(rw->www.author) + p->len + 1);
    len = strlen(rw->www.author);
  }

  strncpy(rw->www.author + len, p->text, p->len);
  rw->www.author[len + p->len] = '\0';

  return;
}

/*
 * ExpandTabs
 */
static char *
ExpandTabs(rw, t, len, tlen, tabexpand)
WWWWidget rw;
char *t;
int len;
int tlen;
int tabexpand;
{
  char *s, *cp, *lastcp, *xs;

  lastcp = t + len;
  s = (char *)WWWGetMem(rw, tlen);
  for (cp = t, xs = s; cp < lastcp; cp++)
  {
    if (*cp == '\t')
    {
      memset(xs, ' ', tabexpand);
      xs += tabexpand;
    }
    else if (*cp != '\r')
    {
      *xs = *cp;
      xs++;
    }
  }

  return(s);
}

/*
 * PlainText
 */
void
PlainText(rw, p, original)
WWWWidget rw;
MLTag *p;
bool original;
{
  char *cp, *lastcp, *start;
  int len;
  XFontStruct *font;
  char *s;
  char *the_text;
  int the_len;
  Boolean tabbed;

  if (original)
  {
    the_text = p->otext;
    the_len = p->olen;
  }
  else
  {
    the_text = p->text;
    the_len = p->len;
  }

  if (the_text == NULL) return;

  font = CurrentFont(rw);

  lastcp = the_text + the_len;
  tabbed = False;
  for (len = 0, start = cp = the_text; cp < lastcp; cp++)
  {
    if (*cp == '\r') continue;
    else if (*cp == '\n')
    {
      if (tabbed)
      {
	s = ExpandTabs(rw, start, cp - start, len, 4);
	CreateTextState(rw, p, s, len, font);
      }
      else
      {
	char *s = WWWGetMem(rw, len);

	memcpy(s, start, len);
	CreateTextState(rw, p, s, len, font);
      }
      AddLineBreak(rw);

      len = 0;
      start = cp + 1;
      tabbed = False;
    }
    else if (*cp == '\t')
    {
      len += 4;
      tabbed = True;
    }
    else len++;
  }
  
  if (tabbed)
  {
    s = ExpandTabs(rw, start, cp - start, len, 4);
    CreateTextState(rw, p, s, len, font);
  }
  else
  {
    char *s = WWWGetMem(rw, len);
    
    memcpy(s, start, len);
    CreateTextState(rw, p, start, len, font);
  }
  
  return;
}

/*
 * FillText
 */
void
FillText(rw, p)
WWWWidget rw;
MLTag *p;
{
  unsigned char *s, *cp, *lastcp;
  char *t;
  XFontStruct *font;
  Boolean found_space;
  RenderPart *rp;

  font = CurrentFont(rw);
  for (lastcp = p->text + p->len, cp = (unsigned char *)p->text;
       cp < lastcp; )
  {
    for (found_space = False; cp < lastcp && isspace(*cp); cp++)
    {
      found_space = True;
    }
    
    if (found_space) rp = CreateTextState(rw, p, NULL, 0, font);
    if (cp == lastcp) break;

    for (s = cp; cp < lastcp && !isspace(*cp); cp++)
	;

    t = (char *)WWWGetMem(rw, cp - s);
    memcpy(t, s, cp - s);
    rp = CreateTextState(rw, p, t, cp - s, font);
  }

  return;
}

void
HandleTitle(rw, end)
WWWWidget rw;
int end;
{
  if (end != 0) DestroyRenderEnv(rw, TAG_TITLE);
  else
  {
    RenderEnv *re;

    re = CreateRenderEnv(rw, TAG_TITLE);
    re->text_mode = TAG_TITLE;
  }

  return;
}

void
HandlePRE(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_PRE);
    AddLineBreak(rw);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(rw);
    re = CreateRenderEnv(rw, TAG_PRE);
    re->font = rw->www.ttFont;
    re->text_mode = TAG_PRE;
    re->wrap = False;
    AddBlankLine(rw);
  }

  return;
}

void
HandleListing(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_LISTING);
    AddLineBreak(rw);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(rw);
    re = CreateRenderEnv(rw, TAG_LISTING);
    re->font = rw->www.ttFont;
    re->text_mode = TAG_LISTING;
    re->wrap = False;
    AddBlankLine(rw);
  }

  return;
}

void
HandlePlaintext(rw, p, end)
WWWWidget rw;
MLTag *p;
int end;
{
  if (end != 0)
  {
    AddBlankLine(rw);
    DestroyRenderEnv(rw, TAG_PLAINTEXT);
    AddLineBreak(rw);
  }
  else
  {
    RenderEnv *re;

    AddLineBreak(rw);
    re = CreateRenderEnv(rw, TAG_PLAINTEXT);
    re->font = rw->www.ttFont;
    re->wrap = False;
    AddBlankLine(rw);
  }

  return;
}



