/*
 * ftp.c
 *
 * Copyright 1993, 1994, 1995 by John Kilburg (john@isri.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"
#include "wwwioextern.h"
#include "net.h"

#define DEFAULT_FTP_PORT 21

typedef struct _ftpinfo
{
  Document *d;
  int datad;
  int controld;
  void (*rfunc)();
  int (*func)();
  void *closure;
  URLParts *up;
  char **sa;
  int sacount;
  int sasize;
  IOContext ioc;
  char dhost[BUFSIZ];
  int dport;
  char *request;
} FTPInfo;

static char *ftp_dir_html _ArgProto((FTPInfo *));

/*
 * ftp_destroy_info
 *
 * Doesn't destroy document.
 */
static void
ftp_destroy_info(fi)
FTPInfo *fi;
{
  int i;

  if (fi->datad > -1) IOEnd(fi->ioc, fi->datad);
  IOEnd(fi->ioc, fi->controld);

  if (fi->sa != NULL)
  {
    for (i = 0; i < fi->sacount; i++)
    {
      if (fi->sa[i] != NULL) free_mem(fi->sa[i]);
    }
    free_mem(fi->sa);
  }

  if (fi->up != NULL) DestroyURLParts(fi->up);
  if (fi->request != NULL) free_mem(fi->request);

  free_mem(fi);

  return;
}

/*
 * ftp_create_info
 */
static FTPInfo *
ftp_create_info()
{
  FTPInfo *fi;

  fi = (FTPInfo *)alloc_mem(sizeof(FTPInfo));
  fi->up = NULL;
  fi->d = NULL;
  fi->func = NULL;
  fi->rfunc = NULL;
  fi->closure = NULL;
  fi->controld = -1;
  fi->datad = -1;
  fi->sa = NULL;
  return(fi);
}

/*
 * ftp_error_message
 *
 * Return error message document.
 */
static Document *
ftp_error_message(ftpmsg)
char *ftpmsg;
{
  char *msg = GetFromStringDB("ftperror");
  char *r;
  Document *d;

  r = alloc_mem(strlen(msg) + strlen(ftpmsg) + 1);
  sprintf (r, msg, ftpmsg);
  free_mem(ftpmsg);
  d = BuildDocument(r, strlen(r), "text/html", 0);

  return(d);
}

/*
 * ftp_error_return
 */
static void
ftp_error_return(fi, t)
FTPInfo *fi;
char *t;
{
  (fi->func)(ftp_error_message(t), fi->closure, DS_FINISHED);

  ftp_destroy_info(fi);

  return;
}

/*
 * ftp_clean
 */
static void
ftp_clean(s, closure, status)
int s;
void *closure;
int status;
{
  FTPInfo *fi = (FTPInfo *)closure;

  if (fi->func != NULL) (fi->func)(fi->d, fi->closure, status);
  ftp_destroy_info(fi);

  return;
}

/*
 * ftp_data
 */
static void
ftp_data(s, t, tlen, eof, closure)
int s;
void *t;
int tlen;
int eof;
void *closure;
{  
  FTPInfo *fi = (FTPInfo *)closure;

  fi->d->data = t;
  fi->d->len = tlen;

  if (eof)
  {
    (fi->func)(fi->d, fi->closure, DS_FINISHED);
    ftp_destroy_info(fi);
  }
  else
  {
    if ((fi->func)(fi->d, fi->closure, DS_DOWNLOADING) < 0)
    {
      ftp_destroy_info(fi);
    }
    else IOReadStream(fi->ioc, fi->datad, ftp_data);
  }

  return;
}

/*
 * ftp_dirdata
 */
static void
ftp_dirdata(s, t, closure)
int s;
char *t;
void *closure;
{  
  FTPInfo *fi = (FTPInfo *)closure;

  if (t == NULL || t[0] == '\0')
  {
    fi->d->data = (void *)ftp_dir_html(fi);
    fi->d->len = strlen((char *)fi->d->data);
    MIMEAddHeaderField(fi->d, "content-type", "text/html");

    (fi->func)(fi->d, fi->closure, DS_FINISHED);

    ftp_destroy_info(fi);
  }
  else
  {
    char *cp;

    if (fi->sa == NULL)
    {
      fi->sacount = 0;
      fi->sasize = 15;
      fi->sa = (char **)alloc_mem(fi->sasize * sizeof(char *));
    }
    else if (fi->sacount >= fi->sasize)
    {
      fi->sasize *= 4;
      fi->sa = (char **)realloc_mem((char *)fi->sa,
				    fi->sasize * sizeof(char *));
    }
    for (cp = t; *cp; cp++)
    {
      if (isspace(*cp)) break;
    }
    fi->sa[fi->sacount] = alloc_mem(cp - t + 1);
    strncpy(fi->sa[fi->sacount], t, cp - t);
    fi->sa[fi->sacount][cp - t] = '\0';
    fi->sacount += 1;
    
    IOReadLine(fi->ioc, fi->datad, ftp_dirdata);
  }

  return;
}

/*
 * ftp_nlst
 */
static void
ftp_nlst(s, t, closure)
int s;
char *t;
void *closure;
{
  FTPInfo *fi = (FTPInfo *)closure;
  int ecode;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_nlst);
  }
  else
  {
    if (ecode >= 400) ftp_error_return(fi, t);
    else IOReadLine(fi->ioc, fi->datad, ftp_dirdata);
  }

  return;
}

/*
 * ftp_write
 */
static void
ftp_write(s, closure)
int s;
void *closure;
{
  FTPInfo *fi = (FTPInfo *)closure;

  if (fi->request != NULL)
  {
    free_mem(fi->request);
    fi->request = NULL;
  }

  IOReadLine(fi->ioc, fi->controld, fi->rfunc);

  return;
}

/*
 * ftp_cwd
 */
static void
ftp_cwd(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_cwd);
    return;
  }

  fi->rfunc = ftp_nlst;
  fi->request = alloc_string("NLST\r\n");
  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write); 

  return;
}

/*
 * ftp_retr
 */
static void
ftp_retr(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_retr);
    return;
  }

  if (ecode < 400)
  {
    char *content;

    content = MIMEExt2Content(fi->up->filename);
    if (content == NULL) content = "text/plain";
    MIMEAddHeaderField(fi->d, "content-type", content);

    IOReadStream(fi->ioc, fi->datad, ftp_data);
  }
  else
  {
    char *format = "CWD %s\r\n";

    fi->rfunc = ftp_cwd;
    fi->request = alloc_mem(strlen(format) + strlen(fi->up->filename) + 1);
    sprintf (fi->request, format, fi->up->filename);

    IOWriteStream(fi->ioc, fi->controld,
		  fi->request, strlen(fi->request), ftp_write);
  }

  return;
}

/*
 * ftp_data_start
 */
static void
ftp_data_start(closure, status)
void *closure;
int status;
{
  FTPInfo *fi = (FTPInfo *)closure;
  char *format = "RETR %s\r\n";

  if (status != DS_DOWNLOADING)
  {
    ftp_destroy_info(fi);
  }
  else
  {
    fi->datad = net_open(fi->dhost, fi->dport);
    if (fi->datad < 0) ftp_error_return(fi, "net_open failed");
    else
    {
      IORegister(fi->ioc, fi->datad, ftp_clean, fi);
      
      fi->request = alloc_mem(strlen(format) + strlen(fi->up->filename) + 1);
      sprintf (fi->request, format, fi->up->filename);
      
      fi->rfunc = ftp_retr;
      IOWriteStream(fi->ioc, fi->controld,
		    fi->request, strlen(fi->request), ftp_write);
    }
  }
  return;
}

/*
 * ftp_transfer
 */
static void
ftp_transfer(s, t, closure)
int s;
char *t;
void *closure;
{
  int h0, h1, h2, h3, p0, p1, reply, n;
  FTPInfo *fi = (FTPInfo *)closure;

  n = sscanf(t, "%d %*[^(] (%d,%d,%d,%d,%d,%d)",
	     &reply, &h0, &h1, &h2, &h3, &p0, &p1);
  if (n != 7 || reply != 227)
  {
    ftp_error_return(fi, "PASV failed");
    return;
  }

  sprintf (fi->dhost, "%d.%d.%d.%d", h0, h1, h2, h3);
  fi->dport = (p0 << 8) + p1;

  IOStart(fi->ioc, ftp_data_start, fi);

  return;
}

/*
 * ftp_pasv
 *
 * Reads the results from the SIZE command.
 * Writes the PASV command.
 */
static void
ftp_pasv(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  char *cp;
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_pasv);
    return;
  }

  fi->d->len = 0;
  if (ecode < 400)
  {
    for (cp = t; *cp; cp++)
    {
      if (*cp == ' ')
      {
	fi->d->len = atoi(cp);
	break;
      }
    }
  }

  fi->request = alloc_string("PASV\r\n");
  fi->rfunc = ftp_transfer;
  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write);

  return;
}

/*
 * ftp_size
 */
static void
ftp_size(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  char *format = "SIZE %s\r\n";
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_size);
    return;
  }

  if (ecode >= 400)
  {
    ftp_error_return(fi, t);
    return;
  }

  fi->request = alloc_mem(strlen(format) + strlen(fi->up->filename) + 1);
  sprintf (fi->request, format, fi->up->filename);
  fi->rfunc = ftp_pasv;

  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write);

  return;
}

/*
 * ftp_type
 */
static void
ftp_type(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_type);
    return;
  }

  if (ecode >= 400)
  {
    ftp_error_return(fi, t);
    return;
  }

  fi->request = alloc_string("TYPE I\r\n");
  fi->rfunc = ftp_size;
  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write);

  return;
}

/*
 * ftp_user
 */
static void
ftp_user(s, t, closure)
int s;
char *t;
void *closure;
{
  FTPInfo *fi = (FTPInfo *)closure;
  int ecode;
  char *uname;
  char *domain;
  static char *format = "PASS -%s@%s\r\n"; 

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_user);
    return;
  }

  if (ecode >= 400)
  {
    ftp_error_return(fi, t);
    return;
  }

  domain = net_gethostname();
  if (domain == NULL)
  {
    ftp_error_return(fi, "could not get hostname");
    return;
  }

  /*
   * Send the password
   */
  if ((uname = getenv("EMAIL")) != NULL)
  {
    fi->request = alloc_mem(strlen(uname) + 9);
    strcpy(fi->request, "PASS -");
    strcat(fi->request, uname);
    strcat(fi->request, "\r\n");
  }
  else
  {
    if ((uname = getenv("USER")) == NULL) uname = "nobody";
    fi->request = alloc_mem(strlen(uname) +
			    strlen(domain) + strlen(format) + 1);
    sprintf(fi->request, format, uname, domain);
  }

  fi->rfunc = ftp_type;
  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write);

  return;
}

/*
 * ftp_greeting
 */
static void
ftp_greeting(s, t, closure)
int s;
char *t;
void *closure;
{
  int ecode;
  FTPInfo *fi = (FTPInfo *)closure;

  if ((ecode = atoi(t)) == 0 || t[3] != ' ')
  {
    IOReadLine(fi->ioc, fi->controld, ftp_greeting);
    return;
  }

  if (ecode >= 400)
  {
    ftp_error_return(fi, t);
    return;
  }

  fi->rfunc = ftp_user;
  fi->request = alloc_string("USER anonymous\r\n");
  IOWriteStream(fi->ioc, fi->controld,
		fi->request, strlen(fi->request), ftp_write);

  return;
}

/*
 * ftp_start
 */
static void
ftp_start(closure, status)
void *closure;
int status;
{
  FTPInfo *fi = (FTPInfo *)closure;
  URLParts *up = fi->up;

  if (status != DS_DOWNLOADING)
  {
    ftp_destroy_info(fi);
  }
  else
  {
    fi->controld = net_open(up->hostname != NULL ? up->hostname:"ftp",
			    (up->port == 0 ? DEFAULT_FTP_PORT:up->port));
    if (fi->controld < 0) ftp_destroy_info(fi);
    else
    {
      IORegister(fi->ioc, fi->controld, ftp_clean, fi);
      
      IOReadLine(fi->ioc, fi->controld, ftp_greeting);
    }
  }

  return;
}

/*
 * ftp
 *
 * Anonymous FTP interface
 *
 */
void
ftp(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  FTPInfo *fi;

  fi = ftp_create_info();
  fi->controld = -1;
  fi->d = CreateDocument();
  fi->up = DupURLParts(up);
  fi->func = func;
  fi->closure = closure;
  fi->sa = NULL;
  fi->ioc = ioc;
  fi->request = NULL;

  IOStart(ioc, ftp_start, fi);

  return;
}

static int
ftp_strcmp(a, b)
char **a, **b;
{
  return(strcmp(*a, *b));
}

/*
 * ftp_dir_html
 *
 * Convert directory list to HTML
 */
static char *
ftp_dir_html(fi)
FTPInfo *fi;
{
  char *f;
  int flen;
  int filelen;
  int entrylen;
  int i;
  char *hostname;
  char *filename;
  char *header = GetFromStringDB("ftpheader");
  static char *entry = "<li> <a href=ftp://%s:%d%s/%s> %s </a>\n";

  hostname = fi->up->hostname;
  filename = fi->up->filename;

  qsort(fi->sa, fi->sacount, sizeof(char *), ftp_strcmp);

  filelen = strlen(filename) + strlen(hostname) + 10;
  entrylen = strlen(entry);

  flen = strlen(header) + 2 * filelen + 1;
  f = (char *)alloc_mem(flen);
  sprintf (f, header, filename, hostname, filename);

  if (filename[0] != '\0' && filename[1] == '\0') filename = "";

  for (i = 0; i < fi->sacount; i++)
  {
    flen += 2 * strlen(fi->sa[i]) + filelen + entrylen;
    f = (char *)realloc_mem(f, flen);
    sprintf(f + strlen(f), entry,
	    hostname,
	    fi->up->port == 0 ? DEFAULT_FTP_PORT:fi->up->port,
	    filename,
	    fi->sa[i], fi->sa[i]);
  }

  return(f);
}
