/*
 * gopher.c
 *
 * Copyright 1993, 1994, 1995 by John Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "common.h"
#include "wwwio.h"
#include "wwwioextern.h"
#include "net.h"

#define DEFAULT_GOPHER_PORT 70

typedef struct _gopherinfo
{
  URLParts *up;
  int (*func)();
  int type;
  void *closure;
  char *filename;
  char *response;
  int rlen;
  IOContext ioc;
} GopherInfo;

/*
 * gopher_destroy_info
 */
static void
gopher_destroy_info(gi)
GopherInfo *gi;
{
  if (gi->filename != NULL) free_mem(gi->filename);
  if (gi->response != NULL) free_mem(gi->response);
  free_mem(gi);

  return;
}

/*
 * gopher_clean
 */
static void
gopher_clean(s, closure, status)
int s;
void *closure;
int status;
{
  GopherInfo *gi = (GopherInfo *)closure;
  Document *d;

  IOEnd(gi->ioc, s);

  d = CreateDocument();

  if (gi->func != NULL) (gi->func)(d, gi->closure, status);

  gopher_destroy_info(gi);

  return;
}

/*
 * gopher_make_document
 */
static void
gopher_make_document(gi)
GopherInfo *gi;
{
  char *content;
  Document *d;
  int type = gi->type;

  if (type == '0') content = "text/plain";
  else if (type == '1') content = "text/html";
  else if (type == '2') content = "text/plain";
  else if (type == '3') content = NULL;
  else if (type == '4') content = "application/mac-binhex40";
  else if (type == '5') content = NULL;
  else if (type == '7') content = "text/html";
  else if (type == '9') content = NULL;
  else if (type == 's') content = "audio/basic";
  else if (type == 'e') content = NULL;
  else if (type == 'I') content = "image/*";
  else if (type == 'M') content = "multipart/mixed";
  else if (type == 'c') content = NULL;
  else if (type == 'g') content = "image/gif";
  else if (type == 'h') content = "text/html";
  else content = NULL;

  d = BuildDocument(gi->response, gi->rlen, content, 0);

  (gi->func)(d, gi->closure, DS_FINISHED);

  gopher_destroy_info(gi);

  return;
}

/*
 * gopher_data
 */
static void
gopher_data(s, t, tlen, eof, closure)
int s;
char *t;
int tlen;
int eof;
void *closure;
{
  GopherInfo *gi = (GopherInfo *)closure;

  IOEnd(gi->ioc, s);

  gi->response = t;
  gi->rlen = tlen;

  gopher_make_document(gi);

  return;
}

#define GOPHER_DISPLAY 0
#define GOPHER_SELECTOR 1
#define GOPHER_HOST 2
#define GOPHER_PORT 3
#define GOPHER_EXTRA 4
#define GOPHER_FIELD_COUNT 5

/*
 * gopher_menu
 *
 * This code scares the hair right off of me.
 *
 * Converts gopher menu lines into HTML.
 *
 * The parts that work right were provided by R. Stewart Ellis many
 * moons ago.  It has been mangled severely to fit the new scheme.
 * 
 */
static void
gopher_menu(s, line, closure)
int s;
char *line;
void *closure;
{
  GopherInfo *gi = (GopherInfo *)closure;
  char gf[GOPHER_FIELD_COUNT][BUFSIZ]; /* fields in gopher line */
  char *form;
  char *junk = NULL;
  char *stuff;
  char *cp;
  char gtype;
  int i;
  int gfno;
  int flen, oflen;
  static char *format  = "<li><a href=gopher://%s:%s/%c%s>%s</a>\n";
  static char *eformat = "<li><a href=gopherp://%s:%s/%c%s>%s</a>\n";
  static char *trailer = "</ul></html>";
  static char *telform = "<li><a href=telnet://%s@%s:%s/>Telnet to %s</a>\n";
  static char *tnform  = "<li><a href=tn3270://%s@%s:%s/>TN3270 to %s</a>\n";

  flen = strlen(gi->response);

  if (line[0] == '\0' || line[0] == '.')
  {
    /*
     * Put a trailer on the HTML document.
     */
    oflen = flen;
    flen += strlen(trailer) + 1;
    gi->response = (char *)realloc_mem(gi->response, flen);
    strcpy(gi->response + oflen, trailer);

    IOEnd(gi->ioc, s);

    gopher_make_document(gi);

    return;
  }

  if ((cp = strchr(line, '\r')) != NULL) *cp = '\0';
  else if ((cp = strchr(line, '\n')) != NULL) *cp = '\0';
  
  /*
   * Make sure that the fields are blank in case someone leaves a
   * field out.
   */
  for (gfno = 0; gfno < GOPHER_FIELD_COUNT; gfno++)
  {
    gf[gfno][0] = '\0';
  }
  
  /*
   * Separate the TAB separated fields.
   */
  gtype = line[0];
  for (i = 0, cp = line + 1, gfno = 0;
       *cp && gfno < GOPHER_FIELD_COUNT;
       cp++)
  {
    if (*cp == '\t')
    {
      gf[gfno][i] = '\0';
      i = 0;
      gfno++;
    }
    else
    {
      gf[gfno][i++] = *cp;
    }
  }
  gf[gfno][i] = '\0';
  
  /*
   * Check for empty fields.
   */
  if (gf[GOPHER_HOST][0] == '\0')
  {
    strcpy(gf[GOPHER_HOST], gi->up->hostname);
  }
  
  /*
   * Select the proper form.
   */
  switch (gtype)
  {
    case '8':
      form = telform;
      stuff = gf[GOPHER_SELECTOR];
      break;
    
    case 'T':
      form = tnform;
      stuff = gf[GOPHER_SELECTOR];
      break;
      
    default:
      if (gtype != '7' &&
	  gf[GOPHER_EXTRA][0] == '+' && gf[GOPHER_SELECTOR][0] != '1')
      {
	form = eformat;
      }
      else
      {
	form = format;
      }
      /*
       * Make sure the weird characters are escaped.
       */
      junk = EscapeURL((unsigned char *)gf[GOPHER_SELECTOR], 0);
      if (junk != NULL) stuff = junk;
      else stuff = gf[GOPHER_SELECTOR];
  }
  
  /*
   * Now make a bit of HTML for the gopher line.
   */
  oflen = flen;
  flen += strlen(form) + strlen(gf[GOPHER_HOST]) + strlen(stuff) +
      strlen(gf[GOPHER_DISPLAY]) + strlen(gf[GOPHER_PORT]) + 2;
  
  gi->response = (char *)realloc_mem(gi->response, flen);
  
  if (gtype != '8' && gtype != 'T')
  {
    sprintf(gi->response + oflen, form,
	    gf[GOPHER_HOST],
	    gf[GOPHER_PORT],    
	    gtype,
	    stuff,
	    gf[GOPHER_DISPLAY]);
  }
  else
  {
    sprintf(gi->response + oflen, form,
	    stuff,
	    gf[GOPHER_HOST],
              gf[GOPHER_PORT],
              gf[GOPHER_DISPLAY]);
  }
  
  if (junk != NULL)
  {
    free_mem(junk);
    junk = NULL;
  }

  IOReadLine(gi->ioc, s, gopher_menu);

  return;
}

/*
 * gopher_write
 *
 * Called when the request is written
 */
static void
gopher_write(s, closure)
int s;
void *closure;
{
  char *stuff;
  GopherInfo *gi = (GopherInfo *)closure;
  static char *header = "<html><title>Gopher directory on %s</title><h1>Gopher directory on %s</h1><ul>\n";
  static char *header2 = "<html><title>Gopher Search Results</title><h1>Gopher Search Results</h1><ul>\n";

  switch(gi->type)
  {
    case '7':
    case '1':
      if (gi->type == '7') gi->response = alloc_string(header2);
      else
      {
	if (gi->up->filename[0] == '\0') stuff = "/";
	else stuff = gi->up->filename;
	
	gi->response = alloc_mem(strlen(header) +
				 strlen(stuff) * 2 +
				 strlen(gi->up->hostname) * 2 + 1);
	sprintf (gi->response, header, stuff,
		 gi->up->hostname, stuff, gi->up->hostname);
      }
      IOReadLine(gi->ioc, s, gopher_menu);
      break;
    default:
      IOReadStream(gi->ioc, s, gopher_data);
  }

  return;
}

/*
 * gopher_start
 */
static void
gopher_start(closure, status)
void *closure;
int status;
{
  GopherInfo *gi = (GopherInfo *)closure;
  URLParts *up = gi->up;
  char *hostname;
  int portno;
  int s;

  if (status != DS_DOWNLOADING)
  {
    gopher_destroy_info(gi);
  }
  else
  {
    if (up->port == 0) portno = DEFAULT_GOPHER_PORT;
    else portno = up->port;
    
    hostname = up->hostname != NULL ? up->hostname:"gopher";
    s = net_open(hostname, portno);
    if (s < 0) gopher_destroy_info(gi);
    else
    {
      IORegister(gi->ioc, s, gopher_clean, gi);
      IOWriteStream(gi->ioc, s,
		    gi->filename, strlen(gi->filename), gopher_write);
    }
  }

  return;
}

/*
 * gopher
 *
 */
void
gopher_main(ioc, up, func, closure, plus)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
int plus;
{
  int filelen;
  int type;
  char *filename;
  char *tfilename = NULL;
  char *ext;
  GopherInfo *gi;

  /*
   * The gopher server WANTS spaces.
   */
  filename = UnescapeURL(up->filename);
  if (filename == NULL) return;

  if (filename[0] == '/' && filename[1] == '\0')
  {
    free_mem(filename);
    filename = alloc_string("");
    type = '1';
  }
  else if (filename[0] == '/')
  {
    type = filename[1];
    tfilename = alloc_string(filename + 2);
    if (filename != NULL) free_mem(filename);
    filename = tfilename;
  }
  else type = filename[0];

  filelen = strlen(filename);

  /*
   * If the type is index then check to see if the user has entered
   * a search string.  If not then return and ask for a query string
   * using the <isindex> command.
   */
  switch (type)
  {
    case '7':
/*
      if (up->input_count == 0)
      {
        char *msg = GetFromStringDB("g<index>");

        d = BuildDocument(msg, strlen(msg), "text/html", 1, 0);
        free_mem(filename);
	(func)(d, closure, DS_FINISHED);
	return;
      }
      else
      {
        if (up->input_values != NULL && up->input_values[0] != NULL)
        {
          ext = up->input_values[0];
        }
        else ext = "";
        tfilename = alloc_mem(filelen + strlen(ext) + 4);
        strcpy(tfilename, filename);
        strcat(tfilename, "\t");
        strcat(tfilename, ext);
        strcat(tfilename, "\r\n");
      }
*/
      break;

    case '8':
    case 'T':
      (func)(NULL, closure, DS_DOWNLOADING);
      break;

    default:
      ext = "";
      tfilename = alloc_mem(filelen + 3);
      strcpy(tfilename, filename);
      strcat(tfilename, "\r\n");
  }

  gi = (GopherInfo *)alloc_mem(sizeof(GopherInfo));
  gi->up = DupURLParts(up);
  gi->func = func;
  gi->type = type;
  gi->closure = closure;
  gi->filename = tfilename;
  gi->ioc = ioc;

  IOStart(ioc, gopher_start, gi);

  return;
}

/*
 * gopherp
 */
void
gopherp(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  gopher_main(ioc, up, func, closure, 1);
  return;
}

/*
 * gopher
 */
void
gopher(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  gopher_main(ioc, up, func, closure, 0);
  return;
}

