/*
 * local.c
 *
 * Copyright (C) 1993, 1994, 1995, John D. Kilburg (john@cs.unlv.edu)
 *
 * See copyright.h for details.
 */
#include "copyright.h"

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#if defined(SYSV) || defined(SVR4) || defined(__arm)
#include <dirent.h>
#define DIRSTUFF struct dirent
#else
#include <sys/dir.h>
#define DIRSTUFF struct direct
#endif

/* Jim Rees fix */
#ifndef S_ISDIR
#define S_ISDIR(mode) (((mode) & S_IFMT) == S_IFDIR)
#endif

#ifndef S_ISBLK
#define S_ISBLK(mode) (((mode) & S_IFMT) == S_IFBLK)
#endif

#ifndef S_ISREG
#define S_ISREG(mode) (((mode) & S_IFMT) == S_IFREG)
#endif

#ifndef S_ISFIFO
#define S_ISFIFO(mode) (((mode) & S_IFMT) == S_IFIFO)
#endif

#ifndef S_ISCHR
#define S_ISCHR(mode) (((mode) & S_IFMT) == S_IFCHR)
#endif

/* not POSIX but what the heck */
#ifndef S_ISLNK
#define S_ISLNK(mode) (((mode) & S_IFMT) == S_IFLNK)
#endif

#define DEFAULT_TELNET_PORT 23

#include "common.h"
#include "wwwio.h"
#include "wwwioextern.h"
#include "net.h"

static Document *LoadFile _ArgProto((char *, int, int *));
static Document *LoadDir _ArgProto((char *, int *));

/*
 * LoadFile
 *
 * Load the contents of a file into memory.
 */
static Document *
LoadFile(filename, size, status)
char *filename;
int size;
int *status;
{
  FILE *fp;
  char *b;
  char *content;
  Document *d;

  fp = fopen(filename, "r");
  if (fp == NULL)
  {
    char *msg = GetFromStringDB("localaccess");

    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    *status = DS_ERROR;
    return(d);
  }

  b = (char *)alloc_mem(size + 1);
  if (fread(b, 1, size, fp) < size)
  {
    char *msg = GetFromStringDB("localaccess");

    free_mem(b);
    fclose(fp);

    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    *status = DS_ERROR;
    return(d);
  }
  b[size] = '\0';

  fclose(fp);

  content = MIMEExt2Content(filename);
  if (content == NULL) content = "text/plain";
  d = BuildDocument(b, size, content, 0);

  *status = DS_FINISHED;

  return(d);
}

static int
local_strcmp(a, b)
char **a, **b;
{
  return(strcmp(*a, *b));
}

/*
 * LoadDir
 *
 * Reads a local directory and converts it into HTML so the user can
 * navigate the local filesystem.
 */
static Document *
LoadDir(filename, status)
char *filename;
int *status;
{
  Document *d;
  DIR *dp;
  DIRSTUFF *de;
  int flen;
  int filelen;
  int formlen;
  int count, size;
  int i;
  char *f;
  struct stat fs;
  char stat_file[BUFSIZ], *cp;
  static char *format = "<li> <a href=file:%s/%s> %s</a>%s\n";
  static char *rformat = "<li> <a href=file:%s%s> %s</a>%s\n";
  char *header = GetFromStringDB("localheader");
  char *useformat;
  char **sa;
  int salen;

  dp = opendir(filename);
  if (dp == NULL)
  {
    char *msg = GetFromStringDB("localaccess");

    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    *status = DS_ERROR;
    return(d);
  }
  
  /*
   * Get the length of various goodies.
   */
  filelen = strlen(filename);
  useformat = (filename[filelen - 1] == '/')? rformat: format;
  formlen = strlen(useformat);
  flen = strlen(header) * 2 + 2 * filelen + 1;

  size = 15;
  sa = (char **)alloc_mem(size * sizeof(char *));

  for (count = 0, de = readdir(dp); de != NULL; de = readdir(dp), count++)
  {
    sprintf (stat_file, "%s/%s", filename, de->d_name);
    if (stat(stat_file, &fs) != -1)
    {
      if (S_ISREG(fs.st_mode))
	  sprintf(stat_file, " (%ld bytes)", (long)fs.st_size);
      else if (S_ISDIR(fs.st_mode)) sprintf(stat_file, "/");
      else if (S_ISLNK(fs.st_mode)) sprintf(stat_file, " &lt;link&gt;");
      else if (S_ISFIFO(fs.st_mode)) sprintf(stat_file, " &lt;pipe&gt;");
      else if (S_ISBLK(fs.st_mode))
	  sprintf(stat_file, " &lt;block device&gt;");
      else if (S_ISCHR(fs.st_mode))
	  sprintf(stat_file, " &lt;char device&gt;");
      else sprintf(stat_file, " &lt;unknown&gt;");
    }
    else strcpy(stat_file, " ");
    
    /*
     * Resize the file entry table if needed.
     */
    if (count >= size)
    {
      size *= 2;
      sa = (char **)realloc_mem((char *)sa, size * sizeof(char *));
    }

    /*
     * Figure out the size of a file entry and for the HTML buffer.
     */
    salen = strlen(de->d_name) + strlen(stat_file) + 2;
    flen += salen * 2 + formlen + filelen;

    /*
     * allocate the memory
     */
    sa[count] = (char *)alloc_mem(salen * sizeof(char));
    sprintf(sa[count], "%s %s", de->d_name, stat_file);
  }

  closedir(dp);

  if (count == 0)
  {
    char *msg = GetFromStringDB("localaccess");

    free_mem((char *)sa);
    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    *status = DS_ERROR;
    return(d);
  }

  qsort(sa, count, sizeof(char *), local_strcmp);

  /*
   * Allocate space for the HTML version.  Hopefully flen is the amount
   * of memory needed to hold the HTML.
   */
  f = (char *)alloc_mem((flen + 1) * sizeof(char));
  sprintf (f, header, filename, filename);

  /*
   * Stuff each file entry into the buffer.
   */
  for (i = 0; i < count; i++)
  {
    for (cp = sa[i]; *cp != '\0'; cp++)
    {
      if (isspace(*cp)) break;
    }
    *cp = '\0';
    ++cp;
    sprintf(f + strlen(f), useformat, filename, sa[i], sa[i], cp); 
    free_mem(sa[i]);
  }
  free_mem((char *)sa);

  *status = DS_FINISHED;

  return(BuildDocument(f, strlen(f), "text/html", 0));
}

/*
 * file
 *
 * Grabs a local file or to grab the file using FTP (if the hostname
 * is not localhost).
 */
void
file(ioc, up, func, closure)
IOContext ioc;
URLParts *up;
int (*func)();
void *closure;
{
  char *domain;
  struct stat s;
  char *filename;
  Document *d;
  char *hostname;
  int status;

  hostname = up->hostname != NULL ? up->hostname:"localhost";
  if (strcasecmp(hostname, "localhost") != 0)
  {
    domain = net_gethostname(); /* gets the full domain name */
    if (domain == NULL) return;
    if (strcasecmp(domain, hostname) != 0)
    {
      char myhostname[256];

      gethostname(myhostname, sizeof(myhostname)); /* gets only hostname ? */
      if (strcasecmp(hostname, myhostname) != 0)
      {
	ftp(ioc, up, func, closure);
	return;
      }
    }
  }

  filename = FixFilename(up->filename);
  if (filename == NULL || stat(filename, &s) < 0)
  {
    char *msg = GetFromStringDB("localaccess");

    d = BuildDocument(msg, strlen(msg), "text/html", 1);
    status = DS_ERROR;
  }
  else
  {
    if (S_ISDIR(s.st_mode)) d = LoadDir(filename, &status);
    else d = LoadFile(filename, (int) s.st_size, &status);
  }

  (func)(d, closure, status);

  return;
}

